package ml2java.parser;

import java.io.Reader;
import java.io.FileReader;
import java.io.IOException;
import java.util.NoSuchElementException;
import java.util.Objects;

import ml2java.util.*;

import static ml2java.util.SimpleList.*;
import static ml2java.util.Pair.*;
import static ml2java.util.Option.*;
import static ml2java.util.Bind.*;

import static ml2java.parser.Exp.*;

public abstract class Comm {
    
    public static Comm Null() {
        return new Null();
    }
    public static Comm Assign(String field1, Exp field2) {
        return new Assign(field1, field2);
    }
    public static Comm Seq(Comm field1, Comm field2) {
        return new Seq(field1, field2);
    }
    public static Comm While(Exp field1, Comm field2) {
        return new While(field1, field2);
    }
    public static Comm IfThen(Exp field1, Comm field2) {
        return new IfThen(field1, field2);
    }
    public static Comm IfThenElse(Exp field1, Comm field2, Comm field3) {
        return new IfThenElse(field1, field2, field3);
    }
    public static Comm CommRegion(Comm field1, Integer field2, Integer field3) {
        return new CommRegion(field1, field2, field3);
    }
    
    public abstract <R> R match(Function0<R> f1, Function2<String, Exp, R> f2, Function2<Comm, Comm, R> f3, Function2<Exp, Comm, R> f4, Function2<Exp, Comm, R> f5, Function3<Exp, Comm, Comm, R> f6, Function3<Comm, Integer, Integer, R> f7);
    
    public abstract <R> R accept(CommVisitor<R> v);
    
    public interface CommVisitor<R> {
    
        public R visitNull();
        public R visitAssign(String s, Exp e);
        public R visitSeq(Comm c1, Comm c2);
        public R visitWhile(Exp e, Comm c);
        public R visitIfThen(Exp e, Comm c);
        public R visitIfThenElse(Exp e, Comm c1, Comm c2);
        public R visitCommRegion(Comm c, Integer i1, Integer i2);
    
    }
    public static abstract class CommDefaultVisitor<R> implements CommVisitor<R> {
    
        public R visitNull() {
            return visitDefault();
        }
        public R visitAssign(String s, Exp e) {
            return visitDefault();
        }
        public R visitSeq(Comm c1, Comm c2) {
            return visitDefault();
        }
        public R visitWhile(Exp e, Comm c) {
            return visitDefault();
        }
        public R visitIfThen(Exp e, Comm c) {
            return visitDefault();
        }
        public R visitIfThenElse(Exp e, Comm c1, Comm c2) {
            return visitDefault();
        }
        public R visitCommRegion(Comm c, Integer i1, Integer i2) {
            return visitDefault();
        }
        public abstract R visitDefault();
    
    }
}

class Null extends Comm { 
    public Null() {
    }
    @Override
    public int hashCode() {
        return Objects.hash();
    }
    @Override
    public boolean equals(Object other) {
        if (other == null) return false;
        return (this.getClass() == (other.getClass()));
    }
    public <R> R match(Function0<R> f1, Function2<String, Exp, R> f2, Function2<Comm, Comm, R> f3, Function2<Exp, Comm, R> f4, Function2<Exp, Comm, R> f5, Function3<Exp, Comm, Comm, R> f6, Function3<Comm, Integer, Integer, R> f7) {
        return f1.apply();
    }
    public <R> R accept(CommVisitor<R> v) {
        return v.visitNull();
    }
}
class Assign extends Comm { 
    final String field1;
    final Exp field2;
    public Assign(String field1, Exp field2) {
        this.field1 = field1;
        this.field2 = field2;
    }
    @Override
    public int hashCode() {
        return Objects.hash(field1, field2);
    }
    @Override
    public boolean equals(Object other) {
        if (other == null) return false;
        return (this.getClass() == (other.getClass())
                    && this.field1.equals(((Assign)other).field1)
                    && this.field2.equals(((Assign)other).field2));
    }
    public <R> R match(Function0<R> f1, Function2<String, Exp, R> f2, Function2<Comm, Comm, R> f3, Function2<Exp, Comm, R> f4, Function2<Exp, Comm, R> f5, Function3<Exp, Comm, Comm, R> f6, Function3<Comm, Integer, Integer, R> f7) {
        return f2.apply(field1, field2);
    }
    public <R> R accept(CommVisitor<R> v) {
        return v.visitAssign(field1, field2);
    }
}
class Seq extends Comm { 
    final Comm field1;
    final Comm field2;
    public Seq(Comm field1, Comm field2) {
        this.field1 = field1;
        this.field2 = field2;
    }
    @Override
    public int hashCode() {
        return Objects.hash(field1, field2);
    }
    @Override
    public boolean equals(Object other) {
        if (other == null) return false;
        return (this.getClass() == (other.getClass())
                    && this.field1.equals(((Seq)other).field1)
                    && this.field2.equals(((Seq)other).field2));
    }
    public <R> R match(Function0<R> f1, Function2<String, Exp, R> f2, Function2<Comm, Comm, R> f3, Function2<Exp, Comm, R> f4, Function2<Exp, Comm, R> f5, Function3<Exp, Comm, Comm, R> f6, Function3<Comm, Integer, Integer, R> f7) {
        return f3.apply(field1, field2);
    }
    public <R> R accept(CommVisitor<R> v) {
        return v.visitSeq(field1, field2);
    }
}
class While extends Comm { 
    final Exp field1;
    final Comm field2;
    public While(Exp field1, Comm field2) {
        this.field1 = field1;
        this.field2 = field2;
    }
    @Override
    public int hashCode() {
        return Objects.hash(field1, field2);
    }
    @Override
    public boolean equals(Object other) {
        if (other == null) return false;
        return (this.getClass() == (other.getClass())
                    && this.field1.equals(((While)other).field1)
                    && this.field2.equals(((While)other).field2));
    }
    public <R> R match(Function0<R> f1, Function2<String, Exp, R> f2, Function2<Comm, Comm, R> f3, Function2<Exp, Comm, R> f4, Function2<Exp, Comm, R> f5, Function3<Exp, Comm, Comm, R> f6, Function3<Comm, Integer, Integer, R> f7) {
        return f4.apply(field1, field2);
    }
    public <R> R accept(CommVisitor<R> v) {
        return v.visitWhile(field1, field2);
    }
}
class IfThen extends Comm { 
    final Exp field1;
    final Comm field2;
    public IfThen(Exp field1, Comm field2) {
        this.field1 = field1;
        this.field2 = field2;
    }
    @Override
    public int hashCode() {
        return Objects.hash(field1, field2);
    }
    @Override
    public boolean equals(Object other) {
        if (other == null) return false;
        return (this.getClass() == (other.getClass())
                    && this.field1.equals(((IfThen)other).field1)
                    && this.field2.equals(((IfThen)other).field2));
    }
    public <R> R match(Function0<R> f1, Function2<String, Exp, R> f2, Function2<Comm, Comm, R> f3, Function2<Exp, Comm, R> f4, Function2<Exp, Comm, R> f5, Function3<Exp, Comm, Comm, R> f6, Function3<Comm, Integer, Integer, R> f7) {
        return f5.apply(field1, field2);
    }
    public <R> R accept(CommVisitor<R> v) {
        return v.visitIfThen(field1, field2);
    }
}
class IfThenElse extends Comm { 
    final Exp field1;
    final Comm field2;
    final Comm field3;
    public IfThenElse(Exp field1, Comm field2, Comm field3) {
        this.field1 = field1;
        this.field2 = field2;
        this.field3 = field3;
    }
    @Override
    public int hashCode() {
        return Objects.hash(field1, field2, field3);
    }
    @Override
    public boolean equals(Object other) {
        if (other == null) return false;
        return (this.getClass() == (other.getClass())
                    && this.field1.equals(((IfThenElse)other).field1)
                    && this.field2.equals(((IfThenElse)other).field2)
                    && this.field3.equals(((IfThenElse)other).field3));
    }
    public <R> R match(Function0<R> f1, Function2<String, Exp, R> f2, Function2<Comm, Comm, R> f3, Function2<Exp, Comm, R> f4, Function2<Exp, Comm, R> f5, Function3<Exp, Comm, Comm, R> f6, Function3<Comm, Integer, Integer, R> f7) {
        return f6.apply(field1, field2, field3);
    }
    public <R> R accept(CommVisitor<R> v) {
        return v.visitIfThenElse(field1, field2, field3);
    }
}
class CommRegion extends Comm { 
    final Comm field1;
    final Integer field2;
    final Integer field3;
    public CommRegion(Comm field1, Integer field2, Integer field3) {
        this.field1 = field1;
        this.field2 = field2;
        this.field3 = field3;
    }
    @Override
    public int hashCode() {
        return Objects.hash(field1, field2, field3);
    }
    @Override
    public boolean equals(Object other) {
        if (other == null) return false;
        return (this.getClass() == (other.getClass())
                    && this.field1.equals(((CommRegion)other).field1)
                    && this.field2.equals(((CommRegion)other).field2)
                    && this.field3.equals(((CommRegion)other).field3));
    }
    public <R> R match(Function0<R> f1, Function2<String, Exp, R> f2, Function2<Comm, Comm, R> f3, Function2<Exp, Comm, R> f4, Function2<Exp, Comm, R> f5, Function3<Exp, Comm, Comm, R> f6, Function3<Comm, Integer, Integer, R> f7) {
        return f7.apply(field1, field2, field3);
    }
    public <R> R accept(CommVisitor<R> v) {
        return v.visitCommRegion(field1, field2, field3);
    }
}

