package ml2java.parser;

import java.io.Reader;
import java.io.FileReader;
import java.io.IOException;
import java.util.NoSuchElementException;
import java.util.Objects;

import ml2java.util.*;

import static ml2java.util.SimpleList.*;
import static ml2java.util.Pair.*;
import static ml2java.util.Option.*;
import static ml2java.util.Bind.*;

import static ml2java.parser.Util.*;
import static ml2java.parser.Token.*;

public class Lexer {

    public static SimpleList<Pair<String, Token>> symbols = list(Pair("==", T_DBLEQUAL()), Pair("!=", T_NOTEQUAL()), Pair("<=", T_LESSEQ()), Pair(">=", T_GREATEREQ()), Pair("&&", T_AND()), Pair("||", T_OR()), Pair(":", T_COLON()), Pair(",", T_COMMA()), Pair("=", T_EQUAL()), Pair("+", T_PLUS()), Pair("-", T_MINUS()), Pair("*", T_TIMES()), Pair("/", T_SLASH()), Pair(";", T_SEMI()), Pair("<", T_LESS()), Pair(">", T_GREATER()), Pair("(", T_LPAR()), Pair(")", T_RPAR()), Pair("{", T_LCURL()), Pair("}", T_RCURL()), Pair("[", T_LSQUARE()), Pair("]", T_RSQUARE()));
    
    public static String implode(SimpleList<Character> l) {
        return match(l,
            /* Nil */ () -> {
                return "";
            },
            /* Cons */ (h, t) -> {
                return (String.valueOf(h) + implode(t));
            });
    }
    
    public static SimpleList<Character> explode_rec(String s, Integer i) {
        if (i < s.length()) {
            return Cons(s.charAt(i), explode_rec(s, (i + 1)));
        } else {
            return Nil();
        }
    }
    
    public static SimpleList<Character> explode(String s) {
        return explode_rec(s, 0);
    }
    
    public static Boolean is_prefix(SimpleList<Character> l1, SimpleList<Character> l2) {
        return match(l1,
            /* Nil */ () -> {
                return true;
            },
            /* Cons */ (h1, t1) -> {
                return match(l2,
                    /* Nil */ () -> {
                        return false;
                    },
                    /* Cons */ (h2, t2) -> {
                        if (h1.equals(h2)) {
                            return is_prefix(t1, t2);
                        } else {
                            return false;
                        }
                    });
            });
    }
    
    public static SimpleList<Character> find_symbol(SimpleList<Pair<String, Token>> sl, SimpleList<Character> l) {
        return match(sl,
            /* Nil */ () -> {
                throw new NoSuchElementException();
            },
            /* Cons */ (h, tl) -> {
                return h.match((s, _4) -> {
                    SimpleList<Character> cl = explode(s);
                    if (is_prefix(cl, l)) {
                        return cl;
                    } else {
                        return find_symbol(tl, l);
                    }
                });
            });
    }
    
    public static <A> SimpleList<A> discard(Integer n, SimpleList<A> l) {
        return match(l,
            /* Nil */ () -> {
                return Nil();
            },
            /* Cons */ (_3, t) -> {
                if (n.equals(0)) {
                    return l;
                } else {
                    return discard((n - 1), t);
                }
            });
    }
    
    public static Pair<SimpleList<Character>, SimpleList<Character>> cut_symbol(SimpleList<Character> l) {
        SimpleList<Character> cl = find_symbol(symbols, l);
        Integer n = size(cl);
        return Pair(cl, discard(n, l));
    }
    
    public static Token make_symbol(String s) {
        return Util.assoc(symbols, s);
    }
    public static SimpleList<Pair<String, Token>> keywords = list(Pair("int", T_WHILE()), Pair("while", T_WHILE()), Pair("if", T_IF()), Pair("then", T_THEN()), Pair("else", T_ELSE()), Pair("return", T_RETURN()));
    
    public static Option<Token> make_kwd(String s) {
        return Util.find(keywords, s);
    }
    
    public static <A> Pair<SimpleList<A>, SimpleList<A>> splitl_rec(Function1<A, Boolean> p, SimpleList<A> l, SimpleList<A> l2) {
        return match(l2,
            /* Nil */ () -> {
                return Pair(l, Nil());
            },
            /* Cons */ (x, tl) -> {
                if (p.apply(x)) {
                    return splitl_rec(p, concat(l, single(x)), tl);
                } else {
                    return Pair(l, l2);
                }
            });
    }
    
    public static <A> Pair<SimpleList<A>, SimpleList<A>> splitl(Function1<A, Boolean> p, SimpleList<A> l) {
        return splitl_rec(p, Nil(), l);
    }
    
    public static <A> Boolean member(A x, SimpleList<A> l) {
        return match(l,
            /* Nil */ () -> {
                return false;
            },
            /* Cons */ (h, t) -> {
                if (x.equals(h)) {
                    return true;
                } else {
                    return member(x, t);
                }
            });
    }
    
    public static <A> Boolean is_empty(SimpleList<A> l) {
        return match(l,
            /* Nil */ () -> {
                return true;
            },
            /* Cons */ (_2, _1) -> {
                return false;
            });
    }
    public static Function1<Integer, Character> chr = (Integer i) -> { int c = i; return (char)c; };
    public static Function1<Character, Integer> code = (Character c) -> { char i = c; return (int)i; };
    
    public static SimpleList<Character> read_all(Reader strm) {
        try {
            Integer b = strm.read();
            if (b < 0) {
                return Nil();
            } else {
                return Cons(chr.apply(b), read_all(strm));
            }
        } catch (IOException _exn) {
            return Nil();
        }
    }
    public static Function1<Character, Boolean> is_letter = (Character c) -> {
            return ((!(c < 'a')) && (!(c > 'z')));
        };
    public static Function1<Character, Boolean> is_cap = (Character c) -> {
            return ((!(c < 'A')) && (!(c > 'Z')));
        };
    public static Function1<Character, Boolean> is_digit = (Character c) -> {
            return ((!(c < '0')) && (!(c > '9')));
        };
    public static Function1<Character, Boolean> is_punct = (Character c) -> {
            return (((!(c < '!')) && (!(c > '~'))) || c.equals('&'));
        };
    public static Function1<Character, Boolean> is_alpha = (Character c) -> {
            return (is_letter.apply(c) || is_cap.apply(c));
        };
    public static Function1<Character, Boolean> is_alpha_num = (Character c) -> {
            return (c.equals('_') || (is_alpha.apply(c) || is_digit.apply(c)));
        };
    public static Function1<Character, Boolean> is_newline = (Character c) -> {
            return code.apply(c).equals(10);
        };
    
    public static Token scan_num(SimpleList<Character> l) {
        return T_Int(Integer.parseInt(implode(l)));
    }
    
    public static Token scan_symb(SimpleList<Character> l) {
        return make_symbol(implode(l));
    }
    
    public static Token scan_alpha_num(SimpleList<Character> l) {
        String s = implode(l);
        Option<Token> opt = make_kwd(s);
        return opt.match( //
            /* None */ () -> {
                return T_Ident(s);
            },
            /* Some */ (kwd) -> {
                return kwd;
            });
    }
    
    public static SimpleList<Pair<Token, Integer>> scan(Integer line, SimpleList<Character> s) {
        return match(s,
            /* Nil */ () -> {
                return Nil();
            },
            /* Cons */ (c, s1) -> {
                if (is_alpha.apply(c)) {
                    Pair<SimpleList<Character>, SimpleList<Character>> p = splitl(is_alpha_num, s);
                    return p.match((l, s2) -> {
                        return Cons(Pair(scan_alpha_num(l), line), scan(line, s2));
                    });
                } else {
                    if (is_digit.apply(c)) {
                        Pair<SimpleList<Character>, SimpleList<Character>> p = splitl(is_digit, s);
                        return p.match((l, s2) -> {
                            return Cons(Pair(scan_num(l), line), scan(line, s2));
                        });
                    } else {
                        if (is_punct.apply(c)) {
                            Pair<SimpleList<Character>, SimpleList<Character>> p = cut_symbol(s);
                            return p.match((l, s2) -> {
                                return Cons(Pair(scan_symb(l), line), scan(line, s2));
                            });
                        } else {
                            if (is_newline.apply(c)) {
                                return scan((line + 1), s1);
                            } else {
                                return scan(line, s1);
                            }
                        }
                    }
                }
            });
    }
    
    public static SimpleList<Pair<Token, Integer>> lexer(SimpleList<Character> s) {
        return scan(1, s);
    }
    
    public static SimpleList<Pair<Token, Integer>> parseFile(String filename) {
        try {
            SimpleList<Character> content = read_all(new FileReader(filename));
            SimpleList<Pair<Token, Integer>> tokens = lexer(content);
            return tokens;
        } catch (Exception msg) {
            throw new Error(msg);
        }
    }
    
    public static void run(String filename) {
        SimpleList<Pair<Token, Integer>> p = parseFile(filename);
        System.out.print(size(p));
        
    }
    
}
