package ml2java.parser;

import java.io.Reader;
import java.io.FileReader;
import java.io.IOException;
import java.util.NoSuchElementException;
import java.util.Objects;

import ml2java.util.*;

import static ml2java.util.SimpleList.*;
import static ml2java.util.Pair.*;
import static ml2java.util.Option.*;
import static ml2java.util.Bind.*;

import static ml2java.parser.Token.*;
import static ml2java.parser.Lexer.*;
import static ml2java.parser.Util.*;
import static ml2java.parser.Typ.*;
import static ml2java.parser.Operation.*;
import static ml2java.parser.Exp.*;
import static ml2java.parser.Comm.*;

public class Parser {

    
    public static Integer getLine(SimpleList<Pair<Token, Integer>> l) {
        return match(l,
            /* Nil */ () -> {
                throw new Error();
            },
            /* Cons */ (h, _15) -> {
                return h.snd();
            });
    }
    
    public static Integer getLastLine(SimpleList<Pair<Token, Integer>> l) {
        return match(l,
            /* Nil */ () -> {
                throw new Error();
            },
            /* Cons */ (h1, l1) -> {
                return match(l1,
                    /* Nil */ () -> {
                        return h1.snd();
                    },
                    /* Cons */ (_14, _13) -> {
                        return getLastLine(l1);
                    });
            });
    }
    
    public static Comm make_seq(SimpleList<Comm> l) {
        return match(l,
            /* Nil */ () -> {
                return Null();
            },
            /* Cons */ (i, l2) -> {
                return match(l2,
                    /* Nil */ () -> {
                        return i;
                    },
                    /* Cons */ (_12, _11) -> {
                        return Seq(i, make_seq(l2));
                    });
            });
    }
    
    public static SimpleList<Pair<Token, Integer>> parse(Pair<Token, SimpleList<Pair<Token, Integer>>> pair) {
        return pair.match((t, s) -> {
            return match(s,
                /* Nil */ () -> {
                    throw new Error("Unexpected EOF");
                },
                /* Cons */ (h, s2) -> {
                    return h.match((tok, line) -> {
                        if (tok.equals(t)) {
                            return s2;
                        } else {
                            throw new Error("Syntax error line " + String.valueOf(line));
                        }
                    });
                });
        });
    }
    
    public static Pair<Exp, SimpleList<Pair<Token, Integer>>> parse_decision(SimpleList<Pair<Token, Integer>> s) {
        return parse_nested_decision(s).match((e1, s1) -> {
            return match(s1,
                /* Nil */ () -> {
                    throw new Error("Unexpected EOF");
                },
                /* Cons */ (h, s2) -> {
                    return h.match((tok, _10) -> {
                        switch (tok.getClass().getSimpleName()) {
                            case "T_AND" : {
                                return parse_decision(s2).match((e2, s3) -> {
                                    return Pair(Oper(And(), e1, e2), s3);
                                });
                            }
                            case "T_OR" : {
                                return parse_decision(s2).match((e2, s3) -> {
                                    return Pair(Oper(Or(), e1, e2), s3);
                                });
                            }
                            default : {
                                return Pair(e1, s1);
                            }
                        }
                    });
                });
        });
    }
    
    public static Pair<Exp, SimpleList<Pair<Token, Integer>>> parse_nested_decision(SimpleList<Pair<Token, Integer>> s) {
        return match(s,
            /* Nil */ () -> {
                throw new Error("Unexpected EOF");
            },
            /* Cons */ (h, s1) -> {
                return h.match((tok, _9) -> {
                    switch (tok.getClass().getSimpleName()) {
                        case "T_LPAR" : {
                            return parse_decision(s1).match((e, s2) -> {
                                return bind(parse(Pair(T_RPAR(), s2)), s3 -> {
                                    return Pair(e, s3);
                                });
                            });
                        }
                        default : {
                            return parse_condition(s);
                        }
                    }
                });
            });
    }
    
    public static Pair<Exp, SimpleList<Pair<Token, Integer>>> parse_condition(SimpleList<Pair<Token, Integer>> s) {
        return parse_exp(s).match((e1, s1) -> {
            return match(s1,
                /* Nil */ () -> {
                    throw new Error("Unexpected EOF");
                },
                /* Cons */ (h, s2) -> {
                    return h.match((tok, line) -> {
                        switch (tok.getClass().getSimpleName()) {
                            case "T_DBLEQUAL" : {
                                return Pair(Equal(), s2);
                            }
                            case "T_NOTEQUAL" : {
                                return Pair(NotEqual(), s2);
                            }
                            case "T_LESS" : {
                                return Pair(Less(), s2);
                            }
                            case "T_GREATER" : {
                                return Pair(Greater(), s2);
                            }
                            case "T_LESSEQ" : {
                                return Pair(LessEq(), s2);
                            }
                            case "T_GREATEREQ" : {
                                return Pair(GreaterEq(), s2);
                            }
                            default : {
                                throw new Error("Syntax error line " + String.valueOf(line));
                            }
                        }
                    }).match((operation, s3) -> {
                        return parse_exp(s3).match((e2, s4) -> {
                            Exp e = Oper(operation, e1, e2);
                            return Pair(e, s4);
                        });
                    });
                });
        });
    }
    
    public static Pair<Exp, SimpleList<Pair<Token, Integer>>> parse_exp(SimpleList<Pair<Token, Integer>> s) {
        return parse_term(s).match((e1, s1) -> {
            return match(s1,
                /* Nil */ () -> {
                    throw new Error("Unexpected EOF");
                },
                /* Cons */ (h, s2) -> {
                    return h.match((tok, _8) -> {
                        switch (tok.getClass().getSimpleName()) {
                            case "T_PLUS" : {
                                return parse_term(s2).match((e2, s3) -> {
                                    return Pair(Oper(Plus(), e1, e2), s3);
                                });
                            }
                            case "T_MINUS" : {
                                return parse_term(s2).match((e2, s3) -> {
                                    return Pair(Oper(Minus(), e1, e2), s3);
                                });
                            }
                            default : {
                                return Pair(e1, s1);
                            }
                        }
                    });
                });
        });
    }
    
    public static Pair<Exp, SimpleList<Pair<Token, Integer>>> parse_term(SimpleList<Pair<Token, Integer>> s) {
        return parse_factor(s).match((e1, s1) -> {
            return match(s1,
                /* Nil */ () -> {
                    throw new Error("Unexpected EOF");
                },
                /* Cons */ (h, s2) -> {
                    return h.match((tok, _7) -> {
                        switch (tok.getClass().getSimpleName()) {
                            case "T_TIMES" : {
                                return parse_factor(s2).match((e2, s3) -> {
                                    return Pair(Oper(Times(), e1, e2), s3);
                                });
                            }
                            case "T_SLASH" : {
                                return parse_factor(s2).match((e2, s3) -> {
                                    return Pair(Oper(Slash(), e1, e2), s3);
                                });
                            }
                            default : {
                                return Pair(e1, s1);
                            }
                        }
                    });
                });
        });
    }
    
    public static Pair<Exp, SimpleList<Pair<Token, Integer>>> parse_factor(SimpleList<Pair<Token, Integer>> s) {
        return match(s,
            /* Nil */ () -> {
                throw new Error("Unexpected EOF");
            },
            /* Cons */ (h, s1) -> {
                return h.match((tok, line) -> {
                    switch (tok.getClass().getSimpleName()) {
                        case "T_Ident" : {
                            return bind(((T_Ident)tok).field1, v -> {
                                return Pair(Var(v), s1);
                            });
                        }
                        case "T_Int" : {
                            return bind(((T_Int)tok).field1, i -> {
                                return Pair(IntCst(i), s1);
                            });
                        }
                        case "T_LPAR" : {
                            return parse_exp(s1).match((e1, s2) -> {
                                return bind(parse(Pair(T_RPAR(), s2)), s3 -> {
                                    return Pair(e1, s3);
                                });
                            });
                        }
                        default : {
                            throw new Error("Syntax error line " + String.valueOf(line));
                        }
                    }
                });
            });
    }
    
    public static Pair<Comm, SimpleList<Pair<Token, Integer>>> parseComm(SimpleList<Pair<Token, Integer>> s) {
        Pair<Comm, SimpleList<Pair<Token, Integer>>> r = parseComm2(s);
        Comm c = r.fst();
        SimpleList<Pair<Token, Integer>> s2 = r.snd();
        return bind(getLine(s), line -> {
            return bind(match(s2,
                /* Nil */ () -> {
                    return getLastLine(s);
                },
                /* Cons */ (_6, _5) -> {
                    return getLine(s2);
                }), line2 -> {
                return Pair(CommRegion(c, line, line2), s2);
            });
        });
    }
    
    public static Pair<Comm, SimpleList<Pair<Token, Integer>>> parseComm2(SimpleList<Pair<Token, Integer>> s) {
        return match(s,
            /* Nil */ () -> {
                throw new Error();
            },
            /* Cons */ (h, s0) -> {
                Token tok = h.fst();
                Integer line = h.snd();
                switch (tok.getClass().getSimpleName()) {
                    case "T_LCURL" : {
                        return parseCommList(s0).match((l, s1) -> {
                            return Pair(make_seq(l), s1);
                        });
                    }
                    case "T_WHILE" : {
                        return bind(parse(Pair(T_LPAR(), s0)), s1 -> {
                            return parse_decision(s1).match((e, s2) -> {
                                return bind(parse(Pair(T_RPAR(), s2)), s3 -> {
                                    return parseComm(s3).match((l, s4) -> {
                                        return Pair(While(e, l), s4);
                                    });
                                });
                            });
                        });
                    }
                    case "T_IF" : {
                        return bind(parse(Pair(T_LPAR(), s0)), s1 -> {
                            return parse_decision(s1).match((e, s2) -> {
                                return bind(parse(Pair(T_RPAR(), s2)), s3 -> {
                                    return parseComm(s3).match((l1, s4) -> {
                                        return match(s4,
                                            /* Nil */ () -> {
                                                throw new Error("Unexpected EOF");
                                            },
                                            /* Cons */ (h2, s5) -> {
                                                return h2.match((tok2, _4) -> {
                                                    switch (tok2.getClass().getSimpleName()) {
                                                        case "T_ELSE" : {
                                                            return parseComm(s5).match((l2, s6) -> {
                                                                return Pair(IfThenElse(e, l1, l2), s6);
                                                            });
                                                        }
                                                        default : {
                                                            return Pair(IfThen(e, l1), s4);
                                                        }
                                                    }
                                                });
                                            });
                                    });
                                });
                            });
                        });
                    }
                    case "T_RETURN" : {
                        return parse_exp(s0).match((e, s1) -> {
                            return Pair(Assign("return", e), s1);
                        });
                    }
                    case "T_Ident" : {
                        return bind(((T_Ident)tok).field1, v -> {
                            return bind(parse(Pair(T_EQUAL(), s0)), s1 -> {
                                return parse_exp(s1).match((e, s2) -> {
                                    return Pair(Assign(v, e), s2);
                                });
                            });
                        });
                    }
                    default : {
                        throw new Error("Syntax error line " + String.valueOf(line));
                    }
                }
            });
    }
    
    public static Pair<SimpleList<Comm>, SimpleList<Pair<Token, Integer>>> parseCommList(SimpleList<Pair<Token, Integer>> s) {
        return match(s,
            /* Nil */ () -> {
                throw new Error("Unexpected EOF");
            },
            /* Cons */ (h, s0) -> {
                return h.match((tok, _3) -> {
                    switch (tok.getClass().getSimpleName()) {
                        case "T_RCURL" : {
                            return Pair(Nil(), s0);
                        }
                        default : {
                            return parseComm(s).match((i, s1) -> {
                                return bind(parse(Pair(T_SEMI(), s1)), s2 -> {
                                    return parseCommList(s2).match((l, s3) -> {
                                        return Pair(Cons(i, l), s3);
                                    });
                                });
                            });
                        }
                    }
                });
            });
    }
    
    public static Comm parseProg(SimpleList<Pair<Token, Integer>> s) {
        return match(s,
            /* Nil */ () -> {
                throw new Error("Unexpected EOF");
            },
            /* Cons */ (h, s0) -> {
                return h.match((tok, _2) -> {
                    switch (tok.getClass().getSimpleName()) {
                        case "T_SEMI" : {
                            return parseCommList(s0).match((l, s1) -> {
                                return match(s1,
                                    /* Nil */ () -> {
                                        return make_seq(l);
                                    },
                                    /* Cons */ (h1, _1) -> {
                                        if (h1.fst().equals(T_RCURL())) {
                                            return make_seq(l);
                                        } else {
                                            throw new Error("EOF expected");
                                        }
                                    });
                            });
                        }
                        default : {
                            return parseProg(s0);
                        }
                    }
                });
            });
    }
    
    public static Comm parseFile(String filename) {
        try {
            SimpleList<Character> content = read_all(new FileReader(filename));
            SimpleList<Pair<Token, Integer>> tokens = Lexer.lexer(content);
            return parseProg(tokens);
        } catch (Exception err) {
            throw new Error(err);
        }
    }
    
    public static void run(String filename) {
        Comm p = parseFile(filename);
        System.out.print("done");
        
    }
    
}
